/* Filename: SRK.c */
/* 17-Apr-2011 */

#include <stdio.h>
#include <math.h>

#define SQ3 1.7320508075688773 /* sqrt(3) */

#define MaxCoreNum 8 /* 8, 1 *//* Maximum number of multi-core. */
/*
  In this program code, OpenMP is available. If you want to use it,
  remove a comment out in the line "#pragma omp parallel for".
 */

extern int OMP_wo2_srk_eff_for_NonCommSDEs_WinMulti(int ydim,
						    unsigned long traj,
						    double *yvec, double step,
						    char *ran3p, char *ran2p,
						    void (*ffunc)(),
						    void (*gfunc_gene)(),
						    int wdim,
						    double work[],
						    double *ynew)
/* This function performs efficient weak second order SRK scheme.
   It gives all trajectries for one step concerning SDEs with
   a multi-dimentional Winer process. If an error occurs, it will
   return 1, otherwise 0.

   Input arguments
   ----------------
   ydim: dimension of SDEs,
   traj: number of trajectries,
   yvec: pointer of the head of all initial values, which are in order
         like yvec[0], yvec[1], ..., yvec[ydim-1] for the 1st trajectry,
	 yvec[ydim], yvec[ydim+1], ..., yvec[2*ydim-1] for the 2nd trajectry.
   step: step length,
   ran3p: pointer of the head of 3*traj three-point distributed RVs
          with P(-1)=P(1)=1/6 and P(0)=2/3,
   ran2p: pointer of the head of 2*traj two-point distributed RVs
          with P(-1)=P(1)=1/2,
   ffunc: drift coefficient,
   gfunc_gene: diffusion coefficients,
   wdim: dimension of Wiener process.

   Workspace arguments
   -------------------
   work: workspace of length MaxCoreNum*(6+wdim*(6+wdim))*ydim.

   Output arguments
   ----------------
   ynew: pointer of the head of all solutions for one step, which are
         in a similar order to yvec.
*/
{
  static int static_flag=0;
  static unsigned long traj_mini;
  int errflag, wdimM1, wdimM2, ii_par;
  /* See Subsection 3.2 in [3] of the references. */
  static double c01=1.0/6, c02=1.0/3, c03=1.0/3, c04=1.0/6,
    cj1=1.0/8, cj2=3.0/8, cj3=3.0/8, cj4=1.0/8,
    /* The coefficients in the deterministic part */
    a0021=1.0/2,
    a0032=1.0/2,
    a0043=1.0,
    /***/
    a0j21=1.0,
    a0j31=-9.0/8, a0j32=9.0/8,
    a0j41=1.0,
    /* The coefficients in the stochastic part */
    ajj21=2.0/3,
    ajj31=1.0/12, ajj32=1.0/4,
    ajj41=-5.0/4, ajj42=1.0/4, ajj43=2.0,
    /***/
    aj021=2.0,
    aj041=-2.0,
    /***/
    ajl31=1.0/4, ajl32=3.0/4,
    ajl41=1.0/4, ajl42=3.0/4,
    /* The coefficients in the stochastic non-commutative part */
    /* See (5) and Section 4. */
    ckj3=-1.0, ckj4, akjjl32, akjjl42;
  double sqstep, sq3step;

  ckj4=-ckj3;
  akjjl32=1.0/(4.0*ckj3);
  akjjl42=-akjjl32;
  
  if((ydim <=0) || (step <= 0))
    {printf("negative argument err\n");return 1;}/* Check for arguments */
  sqstep = sqrt(step);
  sq3step=SQ3*sqstep;
  
  wdimM1=wdim-1;
  wdimM2=wdim-2;

  if (0==static_flag) {
    if (0!=traj%MaxCoreNum) {
      printf("Error: Number of trajects must be %d multiple!", MaxCoreNum);
      exit(1);
    }
    traj_mini=traj/MaxCoreNum;
    static_flag=1;
  }

/*#pragma omp parallel for*/
  for (ii_par=0; ii_par<MaxCoreNum; ii_par++) {
    unsigned long itr;
    int ii, jj, ll, ibase3p, ibase2p, ibase, ibase_work, ibase_work_step;
    double *yn, *yn1, *fn, *Y001, *Y002, *Y003, *Y004, 
      tmp1, tmp2, tmp3, tmp4, tmp5, tmp6;
    double **gn, *dummy, **Yjj1, **Yjl2, **Yjj2, **Yjj3, **Ykj3, **Yjj4,
      **Ykj4;
    char *wj, *wtj;

    if (NULL == (gn = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srk_eff_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (Yjj1 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srk_eff_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    /* This means Y^(j,1), ..., Y^(j,j-1), Y^(j,j+1), ..., Y^(j,wdim) */
    if (NULL == (Yjl2 = (double **)malloc(sizeof(dummy)*wdim*wdimM1))) {
      printf("malloc error in OMP_wo2_srk_eff_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (Yjj2 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srk_eff_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (Yjj3 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srk_eff_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (Ykj3 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srk_eff_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (Yjj4 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srk_eff_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (Ykj4 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srk_eff_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }

    ibase_work_step=6+7*wdim+wdim*wdimM1;
    ibase_work=(ibase_work_step*ydim)*ii_par; /**/
    ii=0;
    yn=&work[ibase_work+ii];
    ii+=ydim;
    fn=&work[ibase_work+ii];
    ii+=ydim;
    Y001=&work[ibase_work+ii];
    ii+=ydim;
    Y002=&work[ibase_work+ii];
    ii+=ydim;
    Y003=&work[ibase_work+ii];
    ii+=ydim;
    Y004=&work[ibase_work+ii];
    ii+=ydim;
    for (jj=0; jj<wdim; jj++) {
      gn[jj]=&work[ibase_work+ii];
      ii+=ydim;
      Yjj1[jj]=&work[ibase_work+ii];
      ii+=ydim;
      Yjj2[jj]=&work[ibase_work+ii];
      ii+=ydim;
      Yjj3[jj]=&work[ibase_work+ii];
      ii+=ydim;
      Ykj3[jj]=&work[ibase_work+ii];
      ii+=ydim;
      Yjj4[jj]=&work[ibase_work+ii];
      ii+=ydim;
      Ykj4[jj]=&work[ibase_work+ii];
      ii+=ydim;
    }
    for (jj=0; jj<wdim*wdimM1; jj++) {
      Yjl2[jj]=&work[ibase_work+ii];
      ii+=ydim;
    }

    ibase=(ydim*traj_mini)*ii_par;
    ibase3p=(wdim*traj_mini)*ii_par;
    ibase2p=(wdimM1*traj_mini)*ii_par;
    for(itr=0; itr<traj_mini; itr++) {
      wj=&ran3p[ibase3p];
      ibase3p+=wdim;
      wtj=&ran2p[ibase2p];
      ibase2p+=wdimM1;
      yn1=&ynew[ibase];
      for(ii=0; ii<ydim; ii++) {
	yn[ii]=yvec[ibase+ii];
      }
      ibase+=ydim;
      ffunc(yn,fn);
      for(ii=0; ii<ydim; ii++) {
	Y001[ii]=step*fn[ii];
      }

      for (jj=0; jj<wdim; jj++) {
	gfunc_gene(yn,jj,gn[jj]);
      }
      for (ll=0; ll<wdim; ll++) {
	switch (wj[ll]) {
	case 0:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj1[ll][ii]=0;
	    for (jj=ll; jj<wdimM1; jj++) {
	      Yjl2[ll*wdimM1+jj][ii]=0;
	      Yjl2[(jj+1)*wdimM1+ll][ii]=0;
	    }
	  }
	  break;
	case 1:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj1[ll][ii]=sq3step*gn[ll][ii];
	  }
	  for (jj=ll; jj<wdimM1; jj++) {
	    if (0<wtj[jj]) {
	      for (ii=0; ii<ydim; ii++) {
		/* This means Y^(l,j)_{s-2} for j>l. */
		/* See also a comment in the part where malloc was called
		   for Yjl2. */
		Yjl2[ll*wdimM1+jj][ii]=sq3step*gn[jj+1][ii];
		/* This means Y^(j,l)_{s-2} for j>l. */
		Yjl2[(jj+1)*wdimM1+ll][ii]=-sq3step*gn[ll][ii];
	      }
	    } else {
	      for (ii=0; ii<ydim; ii++) {
		Yjl2[ll*wdimM1+jj][ii]=-sq3step*gn[jj+1][ii];
		Yjl2[(jj+1)*wdimM1+ll][ii]=sq3step*gn[ll][ii];
	      }
	    }
	  }
	  break;
	default:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj1[ll][ii]=-sq3step*gn[ll][ii];
	  }
	  for (jj=ll; jj<wdimM1; jj++) {
	    if (0<wtj[jj]) {
	      for (ii=0; ii<ydim; ii++) {
		Yjl2[ll*wdimM1+jj][ii]=-sq3step*gn[jj+1][ii];
		Yjl2[(jj+1)*wdimM1+ll][ii]=sq3step*gn[ll][ii];
	      }
	    } else {
	      for (ii=0; ii<ydim; ii++) {
		Yjl2[ll*wdimM1+jj][ii]=sq3step*gn[jj+1][ii];
		Yjl2[(jj+1)*wdimM1+ll][ii]=-sq3step*gn[ll][ii];
	      }
	    }
	  }
	}
      } /* End of the loop for ll*/

      for (ii=0; ii<ydim; ii++) {
	tmp1=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=Yjj1[jj][ii];
	}
	Y002[ii]=yn[ii]+a0021*Y001[ii]+a0j21*tmp1;
      }
      ffunc(Y002,fn);
      for (ii=0; ii<ydim; ii++) {
	Y002[ii]=step*fn[ii];
      }

      for (ii=0; ii<ydim; ii++) {
	tmp1=yn[ii]+aj021*Y001[ii];
	for (jj=0; jj<wdim; jj++) {
	  Yjj2[jj][ii]=tmp1+ajj21*Yjj1[jj][ii];
	}
      }
      for (jj=0; jj<wdim; jj++) {
	gfunc_gene(Yjj2[jj],jj,gn[jj]);
	switch (wj[jj]) {
	case 0:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj2[jj][ii]=0;
	  }
	  break;
	case 1:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj2[jj][ii]=sq3step*gn[jj][ii];
	  }
	  break;
	default:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj2[jj][ii]=-sq3step*gn[jj][ii];
	  }
	}
      }
      
      for (ii=0; ii<ydim; ii++) {
	tmp1=0;
	tmp2=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=Yjj1[jj][ii];
	  tmp2+=Yjj2[jj][ii];
	}
	Y003[ii]=yn[ii]+a0032*Y002[ii]+a0j31*tmp1+a0j32*tmp2;
      }
      ffunc(Y003,fn);
      for (ii=0; ii<ydim; ii++) {
	Y003[ii]=step*fn[ii];
      }

      for (jj=0; jj<wdim; jj++) {
	for (ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  tmp2=0;
	  for (ll=0; ll<wdim; ll++) {
	    if (ll!=jj) {
	      tmp1+=Yjj1[ll][ii];
	      tmp2+=Yjj2[ll][ii];
	    }
	  }
	  Yjj3[jj][ii]=yn[ii]+ajj31*Yjj1[jj][ii]+ajj32*Yjj2[jj][ii]
	    +ajl31*tmp1+ajl32*tmp2;
	}
	gfunc_gene(Yjj3[jj],jj,gn[jj]);
	switch (wj[jj]) {
	case 0:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj3[jj][ii]=0;
	  }
	  break;
	case 1:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj3[jj][ii]=sq3step*gn[jj][ii];
	  }
	  break;
	default:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj3[jj][ii]=-sq3step*gn[jj][ii];
	  }
	}
      }

      for (jj=0; jj<wdim; jj++) {
	for (ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  for (ll=0; ll<wdimM1; ll++) {
	    tmp1+=Yjl2[jj*wdimM1+ll][ii];
	  }
	  Ykj3[jj][ii]=yn[ii]+akjjl32*tmp1;
	}
	gfunc_gene(Ykj3[jj],jj,gn[jj]);
	for (ii=0; ii<ydim; ii++) {
	  Ykj3[jj][ii]=sqstep*gn[jj][ii];
	}
      }
      
      for (ii=0; ii<ydim; ii++) {
	tmp1=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=Yjj1[jj][ii];
	}
	Y004[ii]=yn[ii]+a0043*Y003[ii]+a0j41*tmp1;
      }
      ffunc(Y004,fn);
      for (ii=0; ii<ydim; ii++) {
	Y004[ii]=step*fn[ii];
      }

      for (jj=0; jj<wdim; jj++) {
	for (ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  tmp2=0;
	  for (ll=0; ll<wdim; ll++) {
	    if(ll!=jj) {
	      tmp1+=Yjj1[ll][ii];
	      tmp2+=Yjj2[ll][ii];
	    }
	  }
	  Yjj4[jj][ii]=yn[ii]
	    +ajj41*Yjj1[jj][ii]+ajj42*Yjj2[jj][ii]+ajj43*Yjj3[jj][ii]
	    +aj041*Y001[ii]+ajl41*tmp1+ajl42*tmp2;
	}
	gfunc_gene(Yjj4[jj],jj,gn[jj]);
	switch (wj[jj]) {
	case 0:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj4[jj][ii]=0;
	  }
	  break;
	case 1:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj4[jj][ii]=sq3step*gn[jj][ii];
	  }
	  break;
	default:
	  for (ii=0; ii<ydim; ii++) {
	    Yjj4[jj][ii]=-sq3step*gn[jj][ii];
	  }
	}
      }

      for (jj=0; jj<wdim; jj++) {
	for (ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  for (ll=0; ll<wdimM1; ll++) {
	    tmp1+=Yjl2[jj*wdimM1+ll][ii];
	  }
	  Ykj4[jj][ii]=yn[ii]+akjjl42*tmp1;
	}
	gfunc_gene(Ykj4[jj],jj,gn[jj]);
	for (ii=0; ii<ydim; ii++) {
	  Ykj4[jj][ii]=sqstep*gn[jj][ii];
	}
      }

      for (ii=0; ii<ydim; ii++) {
	tmp1=tmp2=tmp3=tmp4=tmp5=tmp6=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=Yjj1[jj][ii];
	  tmp2+=Yjj2[jj][ii];
	  tmp3+=Yjj3[jj][ii];
	  tmp4+=Yjj4[jj][ii];
	  tmp5+=Ykj3[jj][ii];
	  tmp6+=Ykj4[jj][ii];
	}
	yn1[ii]=yn[ii]+c01*Y001[ii]+c02*Y002[ii]+c03*Y003[ii]+c04*Y004[ii]
	  +cj1*tmp1+cj2*tmp2+cj3*tmp3+cj4*tmp4+ckj3*tmp5+ckj4*tmp6;
      }
    } /* End of loop for itr */
    free(gn);
    free(Yjj1);
    free(Yjl2);
    free(Yjj2);
    free(Yjj3);
    free(Ykj3);
    free(Yjj4);
    free(Ykj4);
  }
  return 0;
}
